from typing import Optional


class ApiException(Exception):

    def __init__(
                self,
                internal_message: str,
                public_message: Optional[str] = None
            ):
        super().__init__(f'{internal_message}: {public_message}')
        self.public_message = public_message



import sys
from argparse import Namespace
from configparser import ConfigParser
from types import SimpleNamespace
from typing import Any, Type, List, Dict, Optional

from .cli_parser import CliCanonicalValueExtractor, get_cli_values, parser
from .config_items import ConfigItemDefinition, \
    AlwaysInvalidExtractor, CanonicalValueExtractorInterface, not_set_token, \
    get_config_map_for_subcommand
from .ini_parser import load_ini, get_ini_value_extractor


class Config(SimpleNamespace):

    def __init__(
                self,
                definitions,
                subcommand,
                ini_path: Optional[str] = None
            ):
        super().__init__()
        self.__definitions = definitions
        self.subcommand = subcommand
        self.ini_path = ini_path
        self.trailing_arguments = None

    def values(self) -> Dict[str, Any]:
        result: Dict[str, Any] = dict()
        for prop, value in vars(self).items():
            if (prop.startswith('_') or callable(value) or
                    isinstance(value, classmethod)):
                continue
            result[prop] = value
        return result

    def get(self, property_name) -> Any:
        return getattr(self, property_name)

    def define(self, property_name) -> ConfigItemDefinition:
        return self.__definitions[property_name]

    def has_ini_file(self) -> bool:
        return self.ini_path is not None


__instance: Optional[Config] = None
__ini_path: Optional[str] = None
__ini_values: Optional[ConfigParser] = None
__cli_values: Optional[Namespace] = None

value_extractors: List = []


def create_config_object(definitions: Dict[str, ConfigItemDefinition],
                         trailing_arguments: List[str], *ordered_sources):
    if len(ordered_sources) < 1:
        raise ValueError("At least one configuration source must be passed in")
    target = Config(definitions, __cli_values.subcommand)
    for source in ordered_sources:
        # if an appropriate extractor isn't found, an exception will be thrown
        extractor_class: Type[
            CanonicalValueExtractorInterface] = AlwaysInvalidExtractor
        for extractor in value_extractors:
            if extractor.is_valid_source(source):
                extractor_class = extractor
                break
        # extract all values from the source and
        # conditionally update the config
        for item_definition in definitions.values():
            new_value = (extractor_class
                         .get_canonical_value(item_definition, source))

            # later values always replace previous values
            if new_value is not not_set_token:
                setattr(target, item_definition.property_name, new_value)
            elif not hasattr(target, item_definition.property_name):
                default = item_definition.default
                if item_definition.has_separator() and \
                        isinstance(default, str):
                    default = default.split(item_definition.meta.separator)
                setattr(target, item_definition.property_name,
                        default)
    target.trailing_arguments = trailing_arguments
    return target


def load_config():
    global __instance
    global __ini_values
    global __cli_values
    if not __instance:
        __cli_values, trailing_arguments = get_cli_values()
        if not __cli_values.subcommand:
            parser.print_help()
            sys.exit()
        __ini_values, __ini_path = load_ini(__cli_values)

        value_extractors.append(get_ini_value_extractor(__cli_values))
        value_extractors.append(CliCanonicalValueExtractor())

        __instance = create_config_object(
            get_config_map_for_subcommand(__cli_values.subcommand),
            trailing_arguments,
            __ini_values,
            __cli_values)
        __instance.ini_path = __ini_path
    return __instance



from logging import Handler
import curses

# based on log handlers in core python logging/__init__.py


class CursesHandler(Handler):

    def __init__(self, window, parent):
        Handler.__init__(self)
        self.window: curses.window = window
        self.parent: curses.window = parent

    def emit(self, record):
        try:
            message = self.format(record)
            self.window.addstr(f"{message}\n")
            self.window.refresh()
            #self.parent.refresh()
        except Exception:
            self.handleError(record)



from ..api.licensing import License, LicenseSpecific


class CommonString:

    def __init__(self, string: str, signature_ids: list = None):
        self.string = string
        if signature_ids is None:
            signature_ids = []
        self.signature_ids = signature_ids


class Signature:

    def __init__(
                self,
                identifier: int,
                rule: str,
                name: str,
                description: str,
                common_strings: list = None
            ):
        self.identifier = identifier
        self.rule = rule
        self.name = name
        self.description = description
        self.common_strings = common_strings \
            if common_strings is not None \
            else []

    def get_common_string_count(self) -> int:
        return len(self.common_strings)

    def has_common_strings(self) -> bool:
        return self.get_common_string_count() > 0


class SignatureSet(LicenseSpecific):

    def __init__(
                self,
                common_strings: list,
                signatures: dict,
                license: License = None
            ):
        super().__init__(license)
        self.common_strings = common_strings
        self.signatures = signatures

    def remove_signature(self, identifier: int) -> bool:
        if identifier not in self.signatures:
            return False
        signature = self.signatures[identifier]
        for index in signature.common_strings:
            self.common_strings[index].signature_ids.remove(identifier)
        del self.signatures[identifier]
        return True

    def get_signature(self, identifier: int) -> None:
        if identifier in self.signatures:
            return self.signatures[identifier]
        raise ValueError(f'Invalid signature identifier: {identifier}')



import importlib
import sys

from .banner.banner import show_welcome_banner_if_enabled
from .config import load_config


def main():
    config = load_config()

    show_welcome_banner_if_enabled(config)

    subcommand_module = importlib.import_module(
            f'.{config.subcommand}.{config.subcommand}',
            package='wordfence.cli'
        )
    exit_code = subcommand_module.main(config)
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



import os
from configparser import ConfigParser, DuplicateSectionError
from multiprocessing import cpu_count

from wordfence.util.input import prompt, prompt_yes_no, prompt_int, \
        InvalidInputException
from wordfence.util.io import ensure_directory_is_writable, resolve_path, \
        IoException
from wordfence.api import noc1
from wordfence.api.licensing import License, LICENSE_URL
from wordfence.api.exceptions import ApiException
from wordfence.logging import log
from ..config.defaults import INI_DEFAULT_PATH


class Configurer:

    def __init__(self, config):
        self.config = config

    def has_base_config(self) -> bool:
        if self.config.license is None:
            return False
        try:
            ensure_directory_is_writable(self.config.cache_directory)
        except IoException:
            log.warning(
                    f'Cache directory at {self.config.cache_directory} does'
                    'not appear to be writable. Please correct the permissions'
                    ' or specify an alternate path for the cache.'
                )
            return False
        return True

    def _prompt_overwrite(self) -> bool:
        if self.config.has_ini_file():
            overwrite = prompt_yes_no(
                    'An existing configuration file was found at '
                    f'{self.config.ini_path}, do you want to overwrite it?',
                    default=False
                )
            return overwrite
        return True

    def _prompt_for_license(self) -> str:
        if self.config.license is None:
            print(f'Please visit {LICENSE_URL} to obtain a license key.')
        else:
            print(f'Current license: {self.config.license}')

        def _validate_license(license: str) -> str:
            client = noc1.Client(License(license), self.config.noc1_url)
            try:
                if not client.ping_api_key():
                    raise InvalidInputException('Invalid license')
            except ApiException as e:
                if e.public_message is not None:
                    raise InvalidInputException(
                            f'Invalid license: {e.public_message}'
                        )
                else:
                    raise InvalidInputException(
                            'License validation failed. Please try again.'
                        )
            return license

        license = prompt(
                'License',
                self.config.license,
                transformer=_validate_license
            )
        return license

    def _prompt_for_cache_directory(self) -> str:

        def _validate_writable(directory: str) -> None:
            try:
                ensure_directory_is_writable(directory)
            except IoException as e:
                raise InvalidInputException(
                        f'Directory {directory} is not writable'
                    ) from e
            return directory

        directory = prompt(
                'Cache directory',
                self.config.cache_directory,
                transformer=_validate_writable
            )
        return directory

    def _prompt_for_worker_count(self) -> int:
        cpus = cpu_count()
        processes = prompt_int(
                    f'Number of worker processes ({cpus} CPUs available)',
                    self.config.workers
                )
        return processes

    def write_config(self) -> None:
        # TODO: What if the INI file changes after the config is loaded?
        config_parser = ConfigParser()
        ini_path = self.config.ini_path if self.config.has_ini_file() \
            else INI_DEFAULT_PATH
        ini_path = resolve_path(ini_path)
        ensure_directory_is_writable(os.path.dirname(ini_path))
        open_mode = 'r' if self.config.has_ini_file() else 'w'
        with open(ini_path, open_mode + '+') as file:
            if self.config.has_ini_file():
                try:
                    config_parser.read_file(file)
                except BaseException:
                    log.warning(
                            'Failed to read existing config file at '
                            f'{ini_path}. existing data will be truncated.'
                        )

            def set_config(key: str, value: str) -> None:
                config_parser.set('SCAN', key, value)

            try:
                config_parser.add_section('SCAN')
            except DuplicateSectionError:
                pass
            set_config('license', self.config.license)
            set_config('cache_directory', self.config.cache_directory)
            set_config('workers', str(self.config.workers))
            file.truncate(0)
            file.seek(0)
            log.info('Writing config...')
            config_parser.write(file)
            log.info('Config updated')

    def prompt_for_config(self) -> None:
        if not self._prompt_overwrite():
            return
        self.config.license = self._prompt_for_license()
        self.config.cache_directory = self._prompt_for_cache_directory()
        self.config.workers = self._prompt_for_worker_count()
        self.write_config()

    def check_config(self) -> None:
        if self.config.configure is False:
            return
        if self.config.configure or not self.has_base_config():
            self.prompt_for_config()



import time


def unit_seconds(ns: int) -> int:
    return ns / 1000000000


class Timer:

    def __init__(self, start: bool = True):
        if start:
            self.start()
        else:
            self.start_time = None
        self.end_time = None

    def _capture_time(self) -> int:
        return time.monotonic_ns()

    def start(self):
        self.start_time = self._capture_time()

    def stop(self):
        self.end_time = self._capture_time()

    def get_elapsed(self, unit=unit_seconds):
        end_time = \
                self.end_time if self.end_time is not None \
                else self._capture_time()
        return unit(end_time - self.start_time)



import re
from typing import Dict, Any

from ..reporting import ReportFormat, ReportColumn
from wordfence.cli.config.defaults import INI_DEFAULT_PATH
from wordfence.util.pcre import PCRE_DEFAULT_MATCH_LIMIT, PCRE_DEFAULT_MATCH_LIMIT_RECURSION

KIBIBYTE = 1024
MEBIBYTE = 1024 * 1024

sizings_map = {
    'b': 1,
    'k': KIBIBYTE,
    'kb': KIBIBYTE,
    'kib': KIBIBYTE,
    'm': MEBIBYTE,
    'mb': MEBIBYTE,
    'mib': MEBIBYTE
}
"""maps suffixes to byte multipliers; k/kb/kib are synonyms, as are m/mb/mib"""


def byte_length(conversion_value: str) -> int:
    match = re.search(r"(\d+)([^0-9].*)", conversion_value)
    if not match:
        raise ValueError("Invalid format for byte length type")
    suffix = match.group(2).lower()
    if not sizings_map.get(suffix, False):
        raise ValueError("Unrecognized byte length suffix")
    return int(match.group(1)) * sizings_map.get(suffix)


config_definitions: Dict[str, Dict[str, Any]] = {
    "read-stdin": {
        "description": "Read paths from stdin. If not specified, paths will "
                       "automatically be read from stdin when input is not "
                       "from a TTY. Specify --no-read-stdin to disable.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "file-list-separator": {
        "short_name": "s",
        "description": "Separator used when listing files via stdin. Defaults "
                       "to the null byte.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "AA==",
        "default_type": "base64"
    },
    "output": {
        "description": "Write results to stdout. This is the default behavior "
                       "when --output-path is not specified. Use --no-output "
                       "to disable.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "output-path": {
        "description": "Path to which to write results.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
    },
    "output-columns": {
        "description": ("An ordered, comma-delimited list of columns to"
                        " include in the output. Available columns: "
                        + ReportColumn.get_valid_options_as_string()),
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "filename",
        "meta": {
            "separator": ","
        }
    },
    "output-format": {
        "short_name": "m",
        "description": "Output format used for result data.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": 'csv',
        "meta": {
            "valid_options": ReportFormat.get_valid_options()
        }
    },
    "output-headers": {
        "description": "Whether or not to include column headers in output",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": None
    },
    "exclude-signatures": {
        "short_name": "e",
        "description": "Specify rule IDs to exclude from the scan. Can be "
                       "comma-delimited and/or specified multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    },
    "include-signatures": {
        "short_name": "i",
        "description": "Specify rule IDs to include in the scan. Can be "
                       "comma-delimited and/or specified multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ",",
            "value_type": int
        }
    },
    "images": {
        "description": "Include image files in the scan.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "include-files": {
        "short_name": "n",
        "description": "Only scan filenames that are exact matches. Can be "
                       "used multiple times.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "exclude-files": {
        "short_name": "x",
        "description": "Do not scan filenames that are exact matches. Can be "
                       "used multiple times. Denials take precedence over "
                       "allows.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "include-files-pattern": {
        "short_name": "N",
        "description": "PCRE regex allow pattern. Only matching filenames will"
                       " be scanned.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "exclude-files-pattern": {
        "short_name": "X",
        "description": "PCRE regex deny pattern. Matching filenames will not "
                       "be scanned.",
        "context": "ALL",
        "argument_type": "OPTION_REPEATABLE",
        "default": None,
        "meta": {
            "separator": ","
        }
    },
    "chunk-size": {
        "short_name": "z",
        "description": "Size of file chunks that will be scanned. Use a whole "
                       "number followed by one of the following suffixes: b "
                       "(byte), k (kibibyte), m (mebibyte). Defaults to 3m.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": 3145728,
        "meta": {
            "value_type": byte_length
        }
    },
    "scanned-content-limit": {
        "short_name": "M",
        "description": "The maximum amount of data to scan in each file."
                       " Content beyond this limit will not be scanned."
                       " Defaults to 50 mebibytes. Use a whole number followed"
                       " by one of the following suffixes: b (byte),"
                       " k (kibibyte), m (mebibyte).",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": byte_length('50m'),
        "meta": {
            "value_type": byte_length
        }
    },
    "match-all": {
        "description": "If set, all possible signatures will be checked "
                       "against each scanned file. Otherwise, only the "
                       "first matching signature will be reported",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "pcre-backtrack-limit": {
        "description": "The regex backtracking limit for signature evaluation",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": PCRE_DEFAULT_MATCH_LIMIT,
        "meta": {
            "value_type": int
        }
    },
    "pcre-recursion-limit": {
        "description": "The regex recursion limit for signature evaluation",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": PCRE_DEFAULT_MATCH_LIMIT_RECURSION,
        "meta": {
            "value_type": int
        }
    },
    "workers": {
        "short_name": "w",
        "description": "Number of worker processes used to perform scanning. "
                       "Defaults to 1 worker process.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": 1
    },
    "configuration": {
        "short_name": "c",
        "description": "Path to a configuration INI file to use (defaults to"
                       f" \"{INI_DEFAULT_PATH}\").",
        "context": "CLI",
        "argument_type": "OPTION",
        "default": INI_DEFAULT_PATH
    },
    "license": {
        "short_name": "l",
        "description": "Specify the license to use.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None
    },
    "cache-directory": {
        "description": "A path to use for cache files.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": "~/.cache/wordfence"
    },
    "cache": {
        "description": "Whether or not to enable the cache.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "purge-cache": {
        "description": "Purge any existing values from the cache.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "verbose": {
        "short_name": "v",
        "description": "Enable verbose logging. If not specified, verbose "
                       "logging will be enabled automatically if stdout is a "
                       "TTY. Use --no-verbose to disable.",
        "context": "ALL",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "debug": {
        "short_name": "d",
        "description": "Enable debug logging.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "quiet": {
        "short_name": "q",
        "description": "Suppress all output other than scan results.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": False
    },
    "banner": {
        "description": "Display the Wordfence banner in command output when "
                       "running in a TTY/terminal.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True
    },
    "progress": {
        "description": "Display scan progress in the terminal with a curses "
                       "interface",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "configure": {
        "description": "Interactively configure Wordfence CLI.",
        "context": "CLI",
        "argument_type": "OPTIONAL_FLAG",
        "default": None
    },
    "version": {
        "description": "Display the version of Wordfence CLI.",
        "context": "CLI",
        "argument_type": "FLAG",
        "default": False
    },
    "check-for-update": {
        "description": "Whether or not to run the update check.",
        "context": "ALL",
        "argument_type": "FLAG",
        "default": True,
        "hidden": True
    },
    "noc1-url": {
        "description": "URL to use for accessing the NOC1 API.",
        "context": "ALL",
        "argument_type": "OPTION",
        "default": None,
        "hidden": True
    },
}



from typing import Optional


class ValidationException(Exception):

    def __init__(self, key: list, message: str, value=None):
        self.key = key
        self.value = value
        super().__init__(
                self.get_key_as_string() +
                ': ' +
                message +
                ', received: ' +
                repr(value)
            )

    def get_key_as_string(self) -> str:
        return '.'.join([str(component) for component in self.key])


class Validator:

    def validate(self, data, parent_key: Optional[list] = None):
        pass

    def validate_type(self, key, value, expected_type) -> None:
        if isinstance(expected_type, Validator):
            expected_type.validate(value, key)
        elif not isinstance(value, expected_type):
            raise ValidationException(
                    key,
                    'Value must be of type ' + str(expected_type),
                    value
                )


class DictionaryValidator(Validator):

    def __init__(self, expected: dict = None):
        self.expected = expected

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if parent_key is None:
            parent_key = []
        if not isinstance(data, dict):
            raise ValidationException(
                    parent_key,
                    'Element must be a dictionary',
                    data
                )
        for key, expected_type in self.expected.items():
            aggregate_key = parent_key + [key]
            try:
                value = data[key]
                self.validate_type(aggregate_key, value, expected_type)
            except KeyError:
                raise ValidationException(aggregate_key, 'Key not present')


class ListValidator(Validator):

    def __init__(self, expected):
        self.expected = expected

    def validate(self, data, parent_key: Optional[list] = None) -> None:
        if parent_key is None:
            parent_key = []
        if not isinstance(data, list):
            raise ValidationException(
                    parent_key,
                    'Element must be a list',
                    data
                )
        if isinstance(self.expected, dict):
            for index, expected_type in self.expected.items():
                key = parent_key + [index]
                try:
                    value = data[index]
                    self.validate_type(key, data[index], expected_type)
                except IndexError:
                    raise ValidationException(
                            key,
                            'Index does not exist in list'
                        )
        else:
            for index, value in enumerate(data):
                self.validate_type(parent_key + [index], value, self.expected)



import os
import queue
import time
from ctypes import c_bool, c_uint
from enum import IntEnum
from multiprocessing import Queue, Process, Value
from dataclasses import dataclass
from typing import Set, Optional, Callable, Dict, NamedTuple
from logging import Handler

from .exceptions import ScanningException
from .matching import Matcher, RegexMatcher
from .filtering import FileFilter, filter_any
from ..util import timing
from ..util.io import StreamReader
from ..util.pcre import PcreOptions, PCRE_DEFAULT_OPTIONS, PcreJitStack
from ..intel.signatures import SignatureSet
from ..logging import log, remove_initial_handler

MAX_PENDING_FILES = 1000  # Arbitrary limit
MAX_PENDING_RESULTS = 100
QUEUE_READ_TIMEOUT = 0
DEFAULT_CHUNK_SIZE = 1024 * 1024
FILE_LOCATOR_WORKER_INDEX = 0
"""Used by the file locator process when sending events"""


class ScanConfigurationException(ScanningException):
    pass


class ScanEventType(IntEnum):
    COMPLETED = 0
    FILE_QUEUE_EMPTIED = 1
    FILE_PROCESSED = 2
    EXCEPTION = 3
    FATAL_EXCEPTION = 4
    PROGRESS_UPDATE = 5
    LOG_MESSAGE = 6


class EventQueueLogHandler(Handler):

    def __init__(self, event_queue: Queue, worker_index: int):
        self._event_queue = event_queue
        self._worker_index = worker_index
        Handler.__init__(self)

    def emit(self, record):
        data = {
                'level': record.levelname,
                'message': record.getMessage()
            }
        self._event_queue.put(
            ScanEvent(
                ScanEventType.LOG_MESSAGE,
                data,
                worker_index=self._worker_index
            )
        )


def use_event_queue_log_handler(event_queue: Queue, worker_index: int) -> None:
    handler = EventQueueLogHandler(
            event_queue,
            worker_index
        )
    remove_initial_handler()
    log.addHandler(handler)


@dataclass
class Options:
    paths: Set[str]
    signatures: SignatureSet
    workers: int = 1
    chunk_size: int = DEFAULT_CHUNK_SIZE
    path_source: Optional[StreamReader] = None
    scanned_content_limit: Optional[int] = None
    file_filter: Optional[FileFilter] = None
    match_all: bool = False
    pcre_options: PcreOptions = PCRE_DEFAULT_OPTIONS


class Status(IntEnum):
    LOCATING_FILES = 0
    PROCESSING_FILES = 1
    COMPLETE = 2
    FAILED = 3


class FileLocator:

    def __init__(self, path: str,
                 queue: Queue,
                 file_filter: FileFilter,
                 ):
        self.path = path
        self.queue = queue
        self.file_filter = file_filter
        self.located_count = 0

    def search_directory(self, path: str):
        try:
            contents = os.scandir(path)
            for item in contents:
                if item.is_dir():
                    yield from self.search_directory(item.path)
                elif item.is_file():
                    if not self.file_filter.filter(item.path):
                        continue
                    self.located_count += 1
                    yield item.path
        except OSError as os_error:
            raise ScanningException('Directory search failed') from os_error

    def locate(self):
        # TODO: Handle links and prevent loops
        real_path = os.path.realpath(self.path)
        if os.path.isdir(real_path):
            for path in self.search_directory(real_path):
                log.debug(f'File added to scan queue: {path}')
                self.queue.put(path)
        else:
            self.queue.put(real_path)


class FileLocatorProcess(Process):

    def __init__(
                self,
                input_queue_size: int = 10,
                output_queue_size: int = MAX_PENDING_FILES,
                file_filter: FileFilter = None,
                use_log_events: bool = False,
                event_queue: Optional[Queue] = None
            ):
        self._input_queue = Queue(input_queue_size)
        self.output_queue = Queue(output_queue_size)
        self.file_filter = file_filter \
            if file_filter is not None \
            else FileFilter([filter_any])
        if use_log_events and not event_queue:
            raise ValueError('Using log events requires an event queue')
        self._use_log_events = use_log_events
        self._event_queue = event_queue
        self._path_count = 0
        super().__init__(name='file-locator')

    def add_path(self, path: str):
        self._path_count += 1
        log.info(f'Scanning path: {path}')
        self._input_queue.put(path)

    def finalize_paths(self):
        self._input_queue.put(None)
        if self._path_count < 1:
            raise ScanConfigurationException(
                    'At least one scan path must be specified'
                )

    def get_next_file(self):
        return self.output_queue.get()

    def run(self):
        if self._use_log_events:
            use_event_queue_log_handler(
                    self._event_queue,
                    FILE_LOCATOR_WORKER_INDEX
                )
        try:
            while (path := self._input_queue.get()) is not None:
                locator = FileLocator(
                        path=path,
                        file_filter=self.file_filter,
                        queue=self.output_queue
                    )
                locator.locate()
            self.output_queue.put(None)
        except ScanningException as exception:
            self.output_queue.put(exception)


class ScanEvent:

    # TODO: Define custom (more compact) pickle serialization format for this
    # class as a potential performance improvement

    def __init__(
                self,
                type: int,
                data=None,
                worker_index: Optional[int] = None
            ):
        self.type = type
        self.data = data
        self.worker_index = worker_index


class ScanProgressMonitor(Process):

    def __init__(self, status: Value, event_queue: Queue):
        super().__init__(name='progress-monitor')
        self._event_queue = event_queue
        self._status = status

    def is_scan_running(self) -> bool:
        status = self._status.value
        return status != Status.COMPLETE and status != Status.FAILED

    def run(self):
        while self.is_scan_running():
            time.sleep(0.1)
            self._event_queue.put(ScanEvent(ScanEventType.PROGRESS_UPDATE))


class ScanWorker(Process):

    def __init__(
                self,
                index: int,
                status: Value,
                work_queue: Queue,
                event_queue: Queue,
                matcher: Matcher,
                chunk_size: int = DEFAULT_CHUNK_SIZE,
                scanned_content_limit: Optional[int] = None,
                use_log_events: bool = False
            ):
        self.index = index
        self._status = status
        self._work_queue = work_queue
        self._event_queue = event_queue
        self._matcher = matcher
        self._chunk_size = chunk_size
        self._working = True
        self._scanned_content_limit = scanned_content_limit
        self._use_log_events = use_log_events
        self.complete = Value(c_bool, False)
        super().__init__(name=self._generate_name())

    def _generate_name(self) -> str:
        return 'worker-' + str(self.index)

    def work(self):
        self._working = True
        log.debug(f'Worker {self.index} started, PID:' + str(os.getpid()))
        with PcreJitStack() as jit_stack:
            while self._working:
                try:
                    item = self._work_queue.get(timeout=QUEUE_READ_TIMEOUT)
                    if item is None:
                        self._put_event(ScanEventType.FILE_QUEUE_EMPTIED)
                        self._complete()
                    elif isinstance(item, BaseException):
                        self._put_event(
                                ScanEventType.FATAL_EXCEPTION,
                                {'exception': item}
                            )
                    else:
                        self._process_file(item, jit_stack)
                except queue.Empty:
                    if self._status.value == Status.PROCESSING_FILES:
                        self._complete()

    def _put_event(self, event_type: ScanEventType, data: dict = None) -> None:
        if data is None:
            data = {}
        self._event_queue.put(
                ScanEvent(event_type, data, worker_index=self.index)
            )

    def _complete(self):
        self._working = False
        self.complete.value = True
        self._put_event(ScanEventType.COMPLETED)

    def is_complete(self) -> bool:
        return self.complete.value

    def _get_next_chunk_size(self, length: int) -> int:
        if self._scanned_content_limit is None:
            return self._chunk_size
        elif length >= self._scanned_content_limit:
            return 0
        else:
            return min(self._scanned_content_limit - length, self._chunk_size)

    def _process_file(self, path: str, jit_stack: PcreJitStack):
        try:
            log.debug(f'Processing file: {path}')
            with open(path, mode='rb') as file, \
                    self._matcher.create_context() as context:
                length = 0
                while (chunk_size := self._get_next_chunk_size(length)):
                    chunk = file.read(chunk_size)
                    if not chunk:
                        break
                    first = length == 0
                    length += len(chunk)
                    if context.process_chunk(chunk, first, jit_stack):
                        break
                self._put_event(
                        ScanEventType.FILE_PROCESSED,
                        {
                            'path': path,
                            'length': length,
                            'matches': context.matches,
                            'timeouts': context.timeouts
                        }
                    )
        except OSError as error:
            self._put_event(ScanEventType.EXCEPTION, {'exception': error})

    def run(self):
        if self._use_log_events:
            use_event_queue_log_handler(self._event_queue, self.index)
        self.work()


class ScanResult:

    def __init__(
                self,
                path: str,
                read_length: int,
                matches: Dict[int, str],
                timeouts: Set[int],
                timestamp: float = None
            ):
        self.path = path
        self.read_length = read_length
        self.matches = matches
        self.timeouts = timeouts
        self.timestamp = timestamp if timestamp is not None else time.time()

    def has_matches(self) -> bool:
        return len(self.matches) > 0

    def get_timeout_count(self) -> int:
        return len(self.timeouts)


class ScanMetrics:

    def __init__(self, worker_count: int):
        self.counts = self._initialize_int_metric(worker_count)
        self.bytes = self._initialize_int_metric(worker_count)
        self.matches = self._initialize_int_metric(worker_count)
        self.timeouts = self._initialize_int_metric(worker_count)

    def _initialize_int_metric(self, worker_count: int):
        return [0] * worker_count

    def record_result(self, worker_index: int, result: ScanResult):
        self.counts[worker_index] += 1
        self.bytes[worker_index] += result.read_length
        if result.has_matches():
            self.matches[worker_index] += 1
        self.timeouts[worker_index] += result.get_timeout_count()

    def _aggregate_int_metric(self, metric: list) -> int:
        total = 0
        for value in metric:
            total += value
        return total

    def get_total_count(self) -> int:
        return self._aggregate_int_metric(self.counts)

    def get_total_bytes(self) -> int:
        return self._aggregate_int_metric(self.bytes)

    def get_total_matches(self) -> int:
        return self._aggregate_int_metric(self.matches)

    def get_total_timeouts(self) -> int:
        return self._aggregate_int_metric(self.timeouts)

    def get_int_metric(self, metric: str, worker_index: Optional[int] = None):
        values = getattr(self, metric)
        if worker_index is not None:
            return values[worker_index]
        else:
            return self._aggregate_int_metric(values)


class ScanProgressUpdate:

    def __init__(self, elapsed_time: int, metrics: ScanMetrics):
        self.elapsed_time = elapsed_time
        self.metrics = metrics


ScanResultCallback = Callable[[ScanResult], None]
ProgressReceiverCallback = Callable[[ScanProgressUpdate], None]
ScanFinishedCallback = Callable[[ScanMetrics, timing.Timer], None]


class ScanFinishedMessages(NamedTuple):
    results: str
    timeouts: Optional[str]


def get_scan_finished_messages(
            metrics: ScanMetrics,
            timer: timing.Timer
        ) -> ScanFinishedMessages:
    match_count = metrics.get_total_matches()
    total_count = metrics.get_total_count()
    byte_count = metrics.get_total_bytes()
    elapsed_time = round(timer.get_elapsed())
    timeout_count = metrics.get_total_timeouts()
    timeouts_message = None
    if timeout_count > 0:
        timeouts_message = f'{timeout_count} timeout(s) occurred during scan'
    results_message = (f'Found {match_count} matching file(s) after '
                       f'processing {total_count} file(s) containing '
                       f'{byte_count} byte(s) over {elapsed_time} second(s)')

    return ScanFinishedMessages(results_message, timeouts_message)


def default_scan_finished_handler(
            metrics: ScanMetrics,
            timer: timing.Timer
        ) -> None:
    """Used as the default ScanFinishedCallback"""
    messages = get_scan_finished_messages(metrics, timer)
    if messages.timeouts:
        log.warning(messages.timeouts)
    log.info(messages.results)


class ScanWorkerPool:

    def __init__(
                self,
                size: int,
                work_queue: Queue,
                event_queue: Queue,
                matcher: Matcher,
                metrics: ScanMetrics,
                timer: timing.Timer,
                progress_receiver: Optional[ProgressReceiverCallback] = None,
                chunk_size: int = DEFAULT_CHUNK_SIZE,
                scanned_content_limit: Optional[int] = None,
                use_log_events: bool = False
            ):
        self.size = size
        self._matcher = matcher
        self._work_queue = work_queue
        self._event_queue = event_queue
        self.metrics = metrics
        self._timer = timer
        self._progress_receiver = progress_receiver
        self._chunk_size = chunk_size
        self._scanned_content_limit = scanned_content_limit
        self._started = False
        self._use_log_events = use_log_events

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.stop()
        else:
            self.terminate()

    def has_progress_receiver(self) -> bool:
        return self._progress_receiver is not None

    def _send_progress_update(self) -> None:
        if self.has_progress_receiver():
            update = ScanProgressUpdate(
                    elapsed_time=self._timer.get_elapsed(),
                    metrics=self.metrics
                )
            self._progress_receiver(update)

    def start(self):
        if self._started:
            raise ScanningException('Worker pool has already been started')
        self._status = Value(c_uint, Status.LOCATING_FILES)
        self._workers = []
        if self.has_progress_receiver():
            self._monitor = ScanProgressMonitor(
                    self._status,
                    self._event_queue
                )
            self._monitor.start()
            self._send_progress_update()
        else:
            self._monitor = None
        for i in range(self.size):
            worker = ScanWorker(
                    i,
                    self._status,
                    self._work_queue,
                    self._event_queue,
                    self._matcher,
                    self._chunk_size,
                    self._scanned_content_limit,
                    self._use_log_events
                )
            worker.start()
            self._workers.append(worker)
        self._started = True

    def _assert_started(self):
        if not self._started:
            raise ScanningException('Worker pool has not been started')

    def stop(self):
        self._assert_started()
        for worker in self._workers:
            worker.join()
        if self._monitor is not None:
            self._monitor.join()

    def terminate(self):
        self._assert_started()
        for worker in self._workers:
            worker.terminate()
        if self._monitor is not None:
            self._monitor.terminate()

    def is_complete(self) -> bool:
        self._assert_started()
        for worker in self._workers:
            if not worker.is_complete():
                return False
        return True

    def await_results(
                self,
                result_processor: ScanResultCallback,
            ):
        self._assert_started()
        while True:
            event = self._event_queue.get()
            if event is None:
                log.debug('All workers have completed and all results have '
                          'been processed.')
                self._status.value = Status.COMPLETE
                return
            elif event.type == ScanEventType.COMPLETED:
                if event.worker_index != FILE_LOCATOR_WORKER_INDEX:
                    log.debug(f'Worker {event.worker_index} completed')
                else:
                    log.debug("File locator process exited")
                if self.is_complete():
                    self._event_queue.put(None)
            elif event.type == ScanEventType.FILE_PROCESSED:
                result = ScanResult(
                        event.data['path'],
                        event.data['length'],
                        event.data['matches'],
                        event.data['timeouts']
                    )
                if result.get_timeout_count() > 0:
                    log.warning(
                            'The following signatures timed out while '
                            f'processing {result.path}: ' +
                            ', '.join({str(i) for i in result.timeouts})
                        )
                self.metrics.record_result(
                        event.worker_index,
                        result
                    )
                result_processor(result)
            elif event.type == ScanEventType.FILE_QUEUE_EMPTIED:
                self._status.value = Status.PROCESSING_FILES
            elif event.type == ScanEventType.EXCEPTION:
                log.error(
                        'Exception occurred while processing file: ' +
                        str(event.data['exception'])
                    )
            elif event.type == ScanEventType.FATAL_EXCEPTION:
                self._status.value = Status.FAILED
                self.terminate()
                raise event.data['exception']
            elif event.type == ScanEventType.PROGRESS_UPDATE:
                self._send_progress_update()
            elif event.type == ScanEventType.LOG_MESSAGE:
                message: str = event.data['message']
                method = getattr(log, event.data['level'].lower())
                method(message)

    def is_failed(self) -> bool:
        return self._status.value == Status.FAILED


class Scanner:

    def __init__(self, options: Options):
        self.options = options
        self.failed = 0

    def _handle_worker_error(self, error: Exception):
        self.failed += 1
        raise error

    def scan(
                self,
                result_processor: ScanResultCallback,
                progress_receiver: Optional[ProgressReceiverCallback] = None,
                scan_finished_handler: ScanFinishedCallback =
                default_scan_finished_handler,
                use_log_events: bool = False
            ):
        """Run a scan"""
        timer = timing.Timer()
        event_queue = Queue(MAX_PENDING_RESULTS)
        file_locator_process = FileLocatorProcess(
                file_filter=self.options.file_filter,
                use_log_events=use_log_events,
                event_queue=event_queue if use_log_events else None
            )
        file_locator_process.start()
        for path in self.options.paths:
            file_locator_process.add_path(path)
        worker_count = self.options.workers
        log.debug("Using " + str(worker_count) + " worker(s)...")
        matcher = RegexMatcher(
                    self.options.signatures,
                    match_all=self.options.match_all,
                    pcre_options=self.options.pcre_options
                )
        metrics = ScanMetrics(worker_count)
        with ScanWorkerPool(
                    size=worker_count,
                    work_queue=file_locator_process.output_queue,
                    event_queue=event_queue,
                    matcher=matcher,
                    metrics=metrics,
                    timer=timer,
                    progress_receiver=progress_receiver,
                    chunk_size=self.options.chunk_size,
                    scanned_content_limit=self.options.scanned_content_limit,
                    use_log_events=use_log_events
                ) as worker_pool:
            if self.options.path_source is not None:
                log.debug('Reading input paths...')
                while True:
                    path = self.options.path_source.read_entry()
                    if path is None:
                        break
                    file_locator_process.add_path(path)
            file_locator_process.finalize_paths()
            log.debug('Awaiting results...')
            worker_pool.await_results(result_processor)
        timer.stop()
        scan_finished_handler = scan_finished_handler if scan_finished_handler\
            else default_scan_finished_handler
        scan_finished_handler(metrics, timer)



import csv
from typing import IO, List, Any
from enum import Enum

from wordfence.scanning.scanner import ScanResult
from wordfence.intel.signatures import SignatureSet, Signature


class ReportColumn(str, Enum):
    FILENAME = 'filename'
    SIGNATURE_ID = 'signature_id'
    SIGNATURE_NAME = 'signature_name'
    SIGNATURE_DESCRIPTION = 'signature_description'
    MATCHED_TEXT = 'matched_text'

    def get_valid_options() -> List[str]:
        return [column.value for column in ReportColumn]

    def get_valid_options_as_string(delimiter: str = ', '):
        return delimiter.join(ReportColumn.get_valid_options())


class ReportFormat(str, Enum):
    CSV = 'csv',
    TSV = 'tsv',
    NULL_DELIMITED = 'null-delimited',
    LINE_DELIMITED = 'line-delimited',
    HUMAN = 'human'

    def get_valid_options() -> List[str]:
        return [format.value for format in ReportFormat]


class ReportWriter:

    def __init__(self, target: IO):
        self._target = target
        self.initialize()

    def initialize(self) -> None:
        pass

    def write_row(self, data: List[str]):
        pass

    def allows_headers(self) -> bool:
        return True


class CsvReportWriter(ReportWriter):

    def initialize(self):
        self.writer = csv.writer(self._target, delimiter=self.get_delimiter())

    def get_delimiter(self) -> str:
        return ','

    def write_row(self, data: List[str]) -> None:
        self.writer.writerow(data)


class TsvReportWriter(CsvReportWriter):

    def get_delimiter(self) -> str:
        return '\t'


class SingleColumnWriter(ReportWriter):

    def __init__(self, target: IO, delimiter: str):
        super().__init__(target)
        self.delimiter = delimiter

    def write_row(self, data: List[str]) -> None:
        for index, value in enumerate(data):
            if index > 0:
                raise ValueError(
                    'Only a single column can be written in this format'
                )
            self._target.write(value + self.delimiter)


class HumanReadableWriter(ReportWriter):

    def __init__(self, target: IO, columns: str):
        super().__init__(target)
        self._columns = columns

    def _get_value(data: List[str], column: str) -> str:
        return

    def _map_data_to_dict(self, data: List[str]) -> dict:
        return {
            name: data[index] for index, name in enumerate(self._columns)
        }

    def write_row(self, data: List[str]) -> None:
        values = self._map_data_to_dict(data)
        file = None
        signature_id = None
        if 'filename' in values:
            file = values['filename']
        if 'signature_id' in values:
            signature_id = values['signature_id']
        # TODO: Add more custom messages if desired
        if file is not None:
            if signature_id is not None:
                self._target.write(
                        f"File at {file} matched signature {signature_id}"
                    )
            else:
                self._target.write(
                        f"File {file} matched a signature"
                    )
        else:
            self._target.write(
                    "Match found: " + str(values)
                )
        self._target.write("\n")

    def allows_headers(self) -> bool:
        return False


class Report:

    def __init__(
                self,
                format: ReportFormat,
                columns: List[str],
                signature_set: SignatureSet,
                write_headers: bool = False
            ):
        self.format = format
        self.columns = columns
        self.signature_set = signature_set
        self.write_headers = write_headers
        self.headers_written = False
        self.writers = []

    def _initialize_writer(self, stream: IO) -> ReportWriter:
        if self.format == ReportFormat.CSV:
            return CsvReportWriter(stream)
        elif self.format == ReportFormat.TSV:
            return TsvReportWriter(stream)
        elif self.format == ReportFormat.NULL_DELIMITED:
            return SingleColumnWriter(stream, "\0")
        elif self.format == ReportFormat.LINE_DELIMITED:
            return SingleColumnWriter(stream, "\n")
        elif self.format == ReportFormat.HUMAN:
            return HumanReadableWriter(stream, self.columns)
        else:
            raise ValueError('Unsupported report format: ' + str(self.format))

    def add_target(self, stream: IO) -> None:
        writer = self._initialize_writer(stream)
        self.writers.append(writer)

    def _get_column_value(
                self,
                column: str,
                result: ScanResult,
                signature: Signature,
                match: str
            ) -> Any:
        if column == ReportColumn.FILENAME.value:
            return result.path
        elif column == ReportColumn.SIGNATURE_ID.value:
            return signature.identifier
        elif column == ReportColumn.SIGNATURE_NAME.value:
            return signature.name
        elif column == ReportColumn.SIGNATURE_DESCRIPTION.value:
            return signature.description
        elif column == ReportColumn.MATCHED_TEXT.value:
            return match,
        elif column == 'discovered_at':
            return int(result.timestamp)
        else:
            raise ValueError(f'Unrecognized column: {column}')

    def _format_result(self, result: ScanResult) -> List[List[str]]:
        rows = []
        for signature_id, match in result.matches.items():
            signature = self.signature_set.get_signature(signature_id)
            row = []
            for column in self.columns:
                value = self._get_column_value(
                        column,
                        result,
                        signature,
                        match
                    )
                row.append(value)
            rows.append(row)
        return rows

    def _write_row(self, data: List[str]):
        for writer in self.writers:
            writer.write_row(data)

    def _write_headers(self) -> None:
        if self.headers_written or not self.write_headers:
            return
        for writer in self.writers:
            if writer.allows_headers():
                writer.write_row(self.columns)

    def add_result(self, result: ScanResult) -> None:
        self._write_headers()
        rows = self._format_result(result)
        for row in rows:
            for writer in self.writers:
                writer.write_row(row)

    def has_writers(self) -> bool:
        return len(self.writers) > 0



import curses
import logging
import os
import unicodedata
from math import ceil
from typing import List, Optional, NamedTuple
from logging import Handler
from collections import deque

from wordfence.scanning.scanner import (ScanProgressUpdate, ScanMetrics,
                                        default_scan_finished_handler)
from ..banner.banner import get_welcome_banner
from ...util import timing

_displays = []

METRIC_BOX_WIDTH = 39
"""
Hard-coded width of metric boxes

The actual width taken up will be the hard-coded value +2 to account for the
left and right borders. Each box on the same row will be separated by the
padding value as well.
"""


class NullLogHandler(Handler):

    def __init__(self):
        Handler.__init__(self)

    def emit(self, record):
        pass


class NullStream:

    def write(self, line):
        pass


def reset_terminal() -> None:
    for display in _displays:
        display.end()


def compute_center_offset(width: int, cols=None) -> int:
    if cols is None:
        cols = curses.COLS
    if width > cols:
        return 0
    return int((cols - width) / 2)


def compute_center_offset_str(string: str, cols=None) -> int:
    return compute_center_offset(len(string), cols)


class StreamToWindow:

    def __init__(self, window: curses.window, parent: curses.window):
        self.window = window
        self.parent = parent

    def write(self, line):
        window = self.window
        # strip control characters
        line = "".join(ch for ch in line if unicodedata.category(ch)[0] != "C")
        window.addstr(f"{line}\n")
        window.refresh()


class Box:

    def __init__(
                self,
                parent: Optional[curses.window] = None,
                border: bool = True,
                title: Optional[str] = None
            ):
        self.parent = parent
        self.border = border
        self.title = title
        self.window = None

    def _initialize_window(self, y: int = 0, x: int = 0) -> None:
        height, width = self.compute_size()
        if self.parent is None:
            self.window = curses.newwin(height, width, y, x)
        else:
            self.window = self.parent.subwin(height, width, y, x)
        self._post_initialize_window()

    def _post_initialize_window(self) -> None:
        pass

    def set_position(self, y: int, x: int) -> None:
        if self.window is None:
            self._initialize_window(y, x)
        else:
            try:
                self.window.mvwin(y, x)
            except Exception:
                raise ValueError(f"error moving window: y: {y}, x: {x}; width:"
                                 f" {self.get_width()}; "
                                 f"height: {self.get_height()}")

    def _require_window(self) -> None:
        if self.window is None:
            self._initialize_window()

    def compute_size(self) -> (int, int):
        height = self.get_height()
        width = self.get_width()
        if self.border:
            width += 2
            height += 2
        return (height, width)

    def resize(self) -> None:
        self.window.clear()
        self.window.refresh()
        height, width = self.compute_size()
        self.window.resize(height, width)

    def set_title(self, title: str) -> None:
        self.title = title

    def render(self) -> None:
        self._require_window()
        height, width = self.compute_size()
        if self.border:
            self.window.border()
        if self.title is not None:
            title_length = len(self.title)
            title_offset = 0
            if title_length < width:
                title_offset = int((width - title_length) / 2)
            self.window.addstr(0, title_offset, self.title)
        self.draw_content()

    def get_border_offset(self) -> int:
        return 1 if self.border else 0

    def draw_content(self) -> None:
        pass

    def update(self) -> None:
        self.render()
        self.window.syncup()
        self.window.noutrefresh()


class Metric:

    def __init__(self, label: str, value):
        self.label = label
        self.value = str(value)


class MetricBox(Box):

    def __init__(
                self,
                metrics: List[Metric],
                title: Optional[str] = None,
                parent: Optional[curses.window] = None
            ):
        self.metrics = metrics
        super().__init__(parent, title=title)

    def get_width(self) -> int:
        return METRIC_BOX_WIDTH

    def get_height(self) -> int:
        return len(self.metrics)

    def draw_content(self) -> None:
        width = self.get_width()
        offset = self.get_border_offset()
        for index, metric in enumerate(self.metrics):
            line = index + offset
            self.window.addstr(line, offset, f'{metric.label}:')
            value_offset = offset + width - len(metric.value)
            self.window.addstr(line, value_offset, metric.value)


class BannerBox(Box):

    def __init__(
                self,
                banner,
                parent: Optional[curses.window] = None
            ):
        self.banner = banner
        super().__init__(parent, border=False)

    def get_width(self):
        # take the full width
        # return self.parent.getmaxyx()[1]
        return self.banner.column_count

    def get_height(self):
        return self.banner.row_count

    def draw_content(self):
        offset = self.get_border_offset()
        for index, row in enumerate(self.banner.rows):
            self.window.addstr(index + offset, offset, row)


class LogBox(Box):

    def __init__(
                self,
                columns: int,
                lines: int,
                parent: Optional[curses.window] = None
            ):
        self.columns = columns
        self.lines = lines
        self.messages = deque()
        super().__init__(parent, border=True)

    def get_width(self):
        return self.columns

    def get_height(self):
        return self.lines

    def _post_initialize_window(self) -> None:
        pass

    def draw_content(self) -> None:
        offset = self.get_border_offset()
        line = offset
        for message in self.messages:
            message = message[:self.columns].ljust(self.columns)
            message = "".join(
                    ch for ch in message if unicodedata.category(ch)[0] != "C"
                )
            self.window.addstr(line, offset, message)
            line += 1

    def add_message(self, message: str) -> None:
        self.messages.append(message)
        while len(self.messages) > self.lines:
            self.messages.popleft()
        self.update()


class LogBoxHandler(Handler):

    def __init__(self, log_box: LogBox):
        self.log_box = log_box
        Handler.__init__(self)

    def emit(self, record):
        self.log_box.add_message(record.getMessage())


class LogBoxStream():

    def __init__(self, log_box: LogBox):
        self.log_box = log_box

    def write(self, line):
        self.log_box.add_message(line)


class BoxLayout:

    def __init__(self, lines: int, cols: int, padding: int = 1):
        self.lines = lines
        self.cols = cols
        self.padding = padding
        self.current_line = 0
        self._content = []
        self._unpositioned = []
        self.max_row_width = 0

    def add_box(self, box: Box) -> None:
        self._content.append(box)
        self._unpositioned.append(box)

    def add_break(self) -> None:
        self._content.append(None)
        self._unpositioned.append(None)

    def _position_row(self, row: list) -> list:
        positioned = []
        extra = []
        row_width = 0
        unpadded_row_width = 0
        row_height = 0
        for box in row:
            height, width = box.compute_size()
            required_width = width + self.padding
            if len(positioned) and (
                        len(extra) or
                        row_width + required_width > self.cols
                    ):
                extra.append(box)
            else:
                row_width += required_width
                unpadded_row_width += width
                row_height = max(row_height, height)
                positioned.append((box, height, width))
        box_count = len(positioned)
        padding = int((self.cols - unpadded_row_width) / (box_count + 1))
        padded_width = unpadded_row_width + padding * (box_count + 1)
        x = padding + int((self.cols - padded_width) / 2)
        final_row_width = 0
        previous_padding = 0
        for (box, height, width) in positioned:
            final_row_width += previous_padding
            y = self.current_line + round((row_height - height) / 2)
            box.set_position(y, x)
            x += width + padding
            final_row_width += width
            previous_padding = padding
            box.update()
        self.current_line += row_height + self.padding
        self.max_row_width = max(self.max_row_width, final_row_width)
        return extra

    def position(self, reset: bool = False) -> None:
        if reset:
            self.current_line = 0
        row = []
        items = self._content if reset else self._unpositioned
        for item in items:
            if item is None:
                row = self._position_row(row)
            else:
                row.append(item)
        while len(row):
            row = self._position_row(row)
        self._unpositioned = []


class LayoutValues(NamedTuple):
    rows: int
    cols: int
    metrics_per_row: int
    metric_rows: int
    metric_height: int
    banner_height: int
    last_metric_line: int
    ideal_message_box_height: int


class ProgressDisplay:

    METRICS_PADDING = 1
    METRICS_COUNT = 5
    MIN_MESSAGE_BOX_HEIGHT = 4

    def __init__(self, worker_count: int):
        _displays.append(self)
        self.worker_count = worker_count
        self._setup_curses()

    def _setup_curses(self) -> None:
        self.stdscr = curses.initscr()
        self._setup_colors()
        curses.noecho()
        curses.curs_set(0)
        self.clear()
        self.banner_box = self._initialize_banner()
        self.metric_boxes = self._initialize_metric_boxes()
        self.layout = self._initialize_layout()
        self.log_box = self._initialize_log_box()
        self.refresh()

    def _setup_colors(self) -> None:
        curses.start_color()
        curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
        self.color_brand = curses.color_pair(1)

    def clear(self):
        self.stdscr.clear()

    def refresh(self):
        self.stdscr.noutrefresh()
        curses.doupdate()

    def end_on_input(self):
        self.stdscr.getkey()
        self.end()

    def end(self):
        curses.endwin()
        _displays.remove(self)

    def _initialize_banner(self) -> Optional[BannerBox]:
        banner = get_welcome_banner()
        if banner is None:
            return None
        return BannerBox(banner=banner, parent=self.stdscr)

    def _compute_rate(self, value: int, elapsed_time: float) -> int:
        if elapsed_time > 0:
            return int(value / elapsed_time)
        return 0

    def _get_metrics(
                self,
                update: ScanProgressUpdate,
                worker_index: Optional[int] = None
            ) -> List[Metric]:
        file_count = update.metrics.get_int_metric('counts', worker_index)
        byte_count = update.metrics.get_int_metric('bytes', worker_index)
        match_count = update.metrics.get_int_metric('matches', worker_index)
        file_rate = self._compute_rate(file_count, update.elapsed_time)
        byte_rate = self._compute_rate(byte_count, update.elapsed_time)
        metrics = [
                Metric('Files Processed', file_count),
                Metric('Bytes Processed', byte_count),
                Metric('Matches Found', match_count),
                Metric('Files / Second', file_rate),
                Metric('Bytes / Second', byte_rate)
            ]
        if len(metrics) > self.METRICS_COUNT:
            raise ValueError("Metrics count is out of sync")
        return metrics

    def _initialize_metric_boxes(self) -> List[MetricBox]:
        default_metrics = ScanMetrics(self.worker_count)
        default_update = ScanProgressUpdate(
                elapsed_time=0,
                metrics=default_metrics
            )
        boxes = []
        for index in range(0, self.worker_count + 1):
            if index == 0:
                worker_index = None
                title = 'Summary'
            else:
                worker_index = index - 1
                title = f'Worker {index}'
            box = MetricBox(
                    self._get_metrics(default_update, worker_index),
                    title=title,
                    parent=self.stdscr
                )
            boxes.append(box)
        return boxes

    def _initialize_log_box(self) -> LogBox:
        log_box = LogBox(
                    columns=self.layout.max_row_width - 2,
                    lines=curses.LINES - self.layout.current_line - 2,
                    parent=self.stdscr
                )
        self.layout.add_box(log_box)
        self.layout.position()
        return log_box

    def _initialize_layout(self) -> BoxLayout:
        layout = BoxLayout(curses.LINES, curses.COLS, self.METRICS_PADDING)
        if self.banner_box is not None:
            layout.add_box(self.banner_box)
        for index, box in enumerate(self.metric_boxes):
            layout.add_box(box)
            if index == 0:
                layout.add_break()
        layout.position()
        return layout

    def _display_metrics(self, update: ScanProgressUpdate) -> None:
        for index in range(0, self.worker_count + 1):
            box = self.metric_boxes[index]
            worker_index = None if index == 0 else index - 1
            box.metrics = self._get_metrics(update, worker_index)
            box.update()

    def handle_update(self, update: ScanProgressUpdate) -> None:
        curses.update_lines_cols()
        self._display_metrics(update)
        self.banner_box.render()
        self.refresh()

    @staticmethod
    def metric_boxes_per_row(columns: int, padding: int = METRICS_PADDING):
        per_row = columns // METRIC_BOX_WIDTH
        if per_row == 0:
            return 0
        display_length = (per_row * METRIC_BOX_WIDTH) + (padding * per_row - 1)
        return per_row if display_length <= columns else per_row - 1

    @staticmethod
    def get_layout_values(worker_count: int,
                          banner_height: Optional[int] = None,
                          cols: Optional[int] = None,
                          rows: Optional[int] = None):
        if banner_height is None:
            banner = get_welcome_banner()
            banner_height = len(banner.rows) + 1 if banner is not None else 0
        if cols is None or rows is None:
            _cols, _rows = os.get_terminal_size()
            cols = _cols if cols is None else cols
            rows = _rows if rows is None else rows
        # one line per metric + two for the top and bottom borders
        metric_height = ProgressDisplay.METRICS_COUNT + 2
        metrics_per_row = ProgressDisplay.metric_boxes_per_row(cols)
        metric_rows = ceil(worker_count / metrics_per_row)
        padding = (0 if banner_height == 0 else 3) + (metric_rows - 1)
        last_metric_line = ((metric_height * metric_rows) + banner_height +
                            padding) - 1
        ideal_message_box_height = 0
        top_and_bottom_adjustment = 2
        final_message_spaceholder = 2
        min_message_box_space = (ProgressDisplay.MIN_MESSAGE_BOX_HEIGHT +
                                 top_and_bottom_adjustment)
        if (last_metric_line + min_message_box_space +
                final_message_spaceholder <= rows):
            max_height = (rows - last_metric_line - top_and_bottom_adjustment
                          - final_message_spaceholder)

            ideal_message_box_height = max_height
        return LayoutValues(rows, cols, metrics_per_row, metric_rows,
                            metric_height, banner_height, last_metric_line,
                            ideal_message_box_height)

    def get_log_handler(self) -> logging.Handler:
        return LogBoxHandler(self.log_box)

    def get_output_stream(self) -> StreamToWindow:
        return LogBoxStream(self.log_box)

    def scan_finished_handler(
                self, metrics: ScanMetrics,
                timer: timing.Timer
            ) -> None:
        default_scan_finished_handler(metrics, timer)
        self.log_box.add_message('Scan completed! Press any key to exit.')
